using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using System.Collections;
using DebugTerminal;

namespace ZavrsniRad
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Simulation : Microsoft.Xna.Framework.Game
    {
        #region varijable

        //varijable simulacije

        KeyboardState currentKeyboardState;

        double simulationStopEnergy;
        double entityMaxEnergy;
        double entityOverFeedingPenalty;
        double foodEatingSpeed;
        double entityEatingSpeed;
        bool startedSimulation;
        bool pausedSimulation;
        bool eofTrial;
        bool showCromosome = false;
        double entityEnergy;
        double entityForwardSpeed;
        double entityRotateSpeed; 
        double entityTimeEnergyLoss;
        double visibilityRange;
        float visibilityAngle;
        Random rand;
        int foodNumber;
        double foodEnergy;
        double foodEnergyLoss;
        int populationNumber;
        List<Entity> entitys;
        List<Food> food;
        List<double> inputs;
        Vector2 randomLocation;
        int speedOfSimulation;
        List<double> foodEyeValues = new List<double>();
        List<double> entityEyeValues = new List<double>();
        
        //varijable GA i neuronske mree

        List<int> nnDesign;
        double neuralNetworkMinWeight;
        double neuralNetworkMaxWeight;
        GeneticAlgorithm ga;
        int geneLength;
        double mutationRate;
        List<BitArray[][]> firstParentChromosome;
        List<BitArray[][]> secondParentChromosome;
        Entity firstParentEntity;
        Entity secondParentEntity;

        //varijable grafikog prikaza

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        Texture2D entitetTexture;
        Texture2D foodTexture;
        Texture2D dotTexture;
        SpriteFont font, statisticsFont;
        Vector2 textInitialOffset;
        Vector2 chromosomeOffset;
        Vector2 textOffsetY;
        Vector2 entitetCenterOffset;
        Vector2 foodCenterOffset;
        List<Vector2> eyeDotReferences = new List<Vector2>();
        Vector2 entitetCenter;
        Vector2 foodCenter;
        Vector2 dotCenter;
        float foodScale;
        Texture2D backTexture;
        float chromosomeRectangleWidth;
        
        //varijable statistike simulacije

        int iterationsInSimulation;
        List<string> entityInformationStrings;
        List<string> simulationInformationStrings;
        int currentGeneration;
        int entitysCount;
        double sumOfEnergyConsumed;
        int numberOfEntityCreated;
        double avgEnergy;
        double avgRatio;
        int numOfEntityNrgConsumed; //koliko najboljih jedinki e biti u statistici na lijevoj strani simulacije, gleda se konzumirana energija
        double avgEnergyForBestEntitys;
        
#endregion
        public Simulation()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>    
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            graphics.PreferredBackBufferWidth = 1024;//rezolucija
            graphics.PreferredBackBufferHeight = 768;
            graphics.ToggleFullScreen();
            graphics.ApplyChanges();

            //Inicijalizacija parametara jedinke
            entityEnergy = 250;
            entityMaxEnergy = 2000;
            entityOverFeedingPenalty = 1000;
            entityForwardSpeed = 8;
            entityRotateSpeed = 25;
            entityTimeEnergyLoss = 0.95;
            visibilityRange = 300;
            visibilityAngle = 35;
            foodEatingSpeed = 10;
            entityEatingSpeed = 35;

            //Inicijalizacija varijabli hrane
            foodNumber = 35;
            foodEnergy = 100;
            foodEnergyLoss = 0.1;
            foodScale = 5f;

            //Inicijalizacija varijabli neuronske mree
            nnDesign = new List<int>();
            nnDesign.Add(8); //dodavanjem integera ovoj varijabli definiramo slojeve neuronske mree
            nnDesign.Add(3);
            nnDesign.Add(3);// ako dodamo 8, 8 i 3, imamo neuronsku mreu 8x8x3 
            neuralNetworkMaxWeight =3; 
            neuralNetworkMinWeight = -3; 

            //Inicijalizacija varijabli genetskog algoritma
            populationNumber = 50;
            geneLength = 16;
            mutationRate = 0.01;

            //Inicijalizacija pomonih varijabli simulacije

            simulationStopEnergy = 70 * entityMaxEnergy;
            sumOfEnergyConsumed = 0;
            startedSimulation = false;
            pausedSimulation = false;
            eofTrial = false;
            currentGeneration = 0;
            speedOfSimulation = 5;
            textInitialOffset = new Vector2(16, -30);
            chromosomeOffset = new Vector2(-16, 30);
            textOffsetY = new Vector2(0, 12);
            entitetCenterOffset = new Vector2(0,0);
            foodCenterOffset = new Vector2(0, 0);
            inputs = new List<double>();
            entityInformationStrings = new List<string>();
            simulationInformationStrings = new List<string>();
            iterationsInSimulation = 0;
            chromosomeRectangleWidth = 10f;
            rand = new Random();
            numOfEntityNrgConsumed = 5;
            avgEnergyForBestEntitys = 0;

            
            numberOfEntityCreated = 0;
            avgEnergy = 0.0;
            avgRatio = 0.0;

            base.Initialize();
        }

        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);
            font = Content.Load<SpriteFont>("FontSimulacije");
            statisticsFont = Content.Load<SpriteFont>("StatisticsFont");
            entitetTexture = Content.Load<Texture2D>("Entitet2");
            entitetCenterOffset.X = entitetTexture.Width / 2;
            entitetCenterOffset.Y = entitetTexture.Height / 2;
            foodTexture = Content.Load<Texture2D>("food2_alpha");
            dotTexture = Content.Load<Texture2D>("ref_tocka_alpha");
            foodCenterOffset.X = foodTexture.Width / 2;
            foodCenterOffset.Y = foodTexture.Height / 2;
            entitetCenter = new Vector2(entitetTexture.Width / 2, entitetTexture.Height / 2);
            foodCenter = new Vector2(foodTexture.Width / 2, foodTexture.Height / 2);
            dotCenter = new Vector2(dotTexture.Width / 2, dotTexture.Height / 2);
            entitys = new List<Entity>();
            food = new List<Food>();
            backTexture = new Texture2D(GraphicsDevice, 1, 1);
            backTexture.SetData(new Color[] { Color.Gray });

            //inicijalizacija Terminala - debug konzole
            Terminal.Init(this, spriteBatch, font, GraphicsDevice);
            Terminal.SetSkin(new TerminalSkin(Color.White, Color.Red, new Color[]{Color.Black,Color.Black, Color.Black, Color.Black },
                1f, Color.LightGreen, Color.LightGreen, Color.LightGreen,Color.LightGreen,Color.LightGreen, Color.LightGreen,Color.LightGreen, Color.Green,Color.LightGreen));

        }
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }
        /// <summary>
        /// Metoda koje se uz metodu Draw() naizmjence automatski poziva tokom rada programa.
        /// U ovu metodi se zbog toga svojstva nalazi simulacijska logika.
        /// </summary>
        /// <param name="gameTime"></param>
        protected override void Update(GameTime gameTime)
        {
            iterationsInSimulation = 0;
            currentKeyboardState = Keyboard.GetState();
            if (Terminal.CheckOpen(Keys.OemTilde, currentKeyboardState) == false)
            {
                currentKeyboardState = CheckKeyboardInput(currentKeyboardState);
            }
            if ((!pausedSimulation) && (startedSimulation == true))
            {
                for (int s = 0; s < speedOfSimulation; s++)
                {
                    foreach (Entity e in entitys)
                    {
                        EntityFeeding(e);
                        e.EntityPosition = CalculatePositionAdjustment(e.EntityPosition);
                        e.ThinkAndAct(CalculateInputs(entitys.IndexOf(e), e.EntityPosition, e.AngleOfMovement));
                        e.Energy -= entityTimeEnergyLoss;
                        e.EnergyDepleted += entityTimeEnergyLoss;
                    }
                  
                        entitys.Sort();
                 
                        sumOfEnergyConsumed = 0;
                        avgRatio = 0;
                        foreach (Entity e in entitys)
                        {
                            if (e.EnergyConsumed > simulationStopEnergy) eofTrial = true;
                            sumOfEnergyConsumed += e.EnergyConsumed;
                            avgRatio += e.MeatToVegetableRatio();
                        }
                        avgEnergy = sumOfEnergyConsumed / entitys.Count;
                        avgRatio = avgRatio / entitys.Count;
                        if (eofTrial)
                        {
                            pausedSimulation = true;
                        }
                        
                        firstParentChromosome = entitys[0].Nn.Chromosome;
                        secondParentChromosome = entitys[1].Nn.Chromosome;
                        firstParentEntity = entitys[0];
                        secondParentEntity = entitys[1];
                    }
                    CheckFoodEnergy();
                    CheckEntityEnergy();
                

                if (entitys.Count() == 0)
                {
                    food.Clear();
                    CreateNewFood(foodNumber);
                    CreateNewGeneration(populationNumber, firstParentChromosome, secondParentChromosome);
                    sumOfEnergyConsumed = 0;
                }
            }
            base.Update(gameTime);
        }
        /// <summary>
        /// Metoda zaduena za iscrtavanje grafike
        /// </summary>
        /// <param name="gameTime"></param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Beige);
            float size;
            spriteBatch.Begin();

            foreach(Food f in food)
            {
                size = (float)((f.Energy / foodEnergy) * foodScale);
                spriteBatch.Draw(foodTexture
                                ,f.FoodPosition
                                ,null
                                ,Color.White
                                ,0f
                                ,foodCenter
                                ,size
                                ,SpriteEffects.None
                                ,1f);
            }
            foreach(Entity e in entitys)
            {
                spriteBatch.Draw(entitetTexture
                                ,e.EntityPosition
                                ,null
                                ,CalculateOverlay(e.Energy, entityEnergy, e.MeatToVegetableRatio())
                                , MathHelper.ToRadians((float)e.AngleOfMovement)
                                , entitetCenter
                                , 1f
                                 ,SpriteEffects.None
                                ,1f);
                
                
                
                entityInformationStrings.Add(e.Energy.ToString("0.0"));
                entityInformationStrings.Add(e.EnergyConsumed.ToString("0.0"));
                DrawStrings(entityInformationStrings, spriteBatch, e.EntityPosition + textInitialOffset, font, Color.Black);
                entityInformationStrings.Clear();
            }
            
            if (startedSimulation == false) simulationInformationStrings.Add("Press S to start simulation");
            if (pausedSimulation == true) simulationInformationStrings.Add("SIMULATION PAUSED!");
            simulationInformationStrings.Add("Trenutna generacija: " + currentGeneration.ToString());
            simulationInformationStrings.Add("Suma konzumirane energije:" + sumOfEnergyConsumed.ToString("0.0"));
            simulationInformationStrings.Add("Brzina simulacije: " + speedOfSimulation.ToString() + "X");
            simulationInformationStrings.Add("Broj stvorenih jedinki: " + numberOfEntityCreated.ToString());
            if (pausedSimulation == true)
            {
                simulationInformationStrings.Add("Prosjecno konzumirana energija: " + avgEnergy.ToString("0.0"));
                simulationInformationStrings.Add("Konzumirana energija od jedinki/hrane: " + avgRatio.ToString("0.0000"));
            }
            if (startedSimulation == true)
            {
                avgEnergyForBestEntitys = 0.0;
                if (entitys.Count < numOfEntityNrgConsumed) { numOfEntityNrgConsumed = entitys.Count; }
                for (int i = 0; i < numOfEntityNrgConsumed; i++)
                {
                    simulationInformationStrings.Add(entitys[i].EnergyConsumed.ToString("0.0"));
                    avgEnergyForBestEntitys += entitys[i].EnergyConsumed;
                }
                avgEnergyForBestEntitys = avgEnergyForBestEntitys / numOfEntityNrgConsumed;
                simulationInformationStrings.Add("Prosjek za " + numOfEntityNrgConsumed.ToString() + ":"+ avgEnergyForBestEntitys.ToString("0.0"));
                
            }
            DrawStrings(simulationInformationStrings, spriteBatch, Vector2.Zero, statisticsFont, Color.Red);
            simulationInformationStrings.Clear();

            if ((entitys.Count > 0) && (showCromosome == true)) 
                DrawChromosome(spriteBatch, entitys[0].Nn.Chromosome, entitys[0].EntityPosition + chromosomeOffset, chromosomeRectangleWidth);
            spriteBatch.End();
            Terminal.CheckDraw(false);
            base.Draw(gameTime);
        }

        #region metode simulacije
        public string Start(int _populationNumber, List<int> _nnDesign, int _geneLength, double _mutationRate, double minWeight, double maxWeight)
        {
            try
            {
                if ((_populationNumber >= 2) && (_nnDesign.Count >= 2) && (_geneLength > 1) && (_geneLength < 32))
                {
                    sumOfEnergyConsumed = 0;
                    numberOfEntityCreated = 0;
                    currentGeneration = 0;
                    entitysCount = 0;
                    populationNumber = _populationNumber;
                    nnDesign = _nnDesign;
                    geneLength = _geneLength;
                    neuralNetworkMinWeight = minWeight;
                    neuralNetworkMaxWeight = maxWeight;
                    food.Clear();
                    entitys.Clear();
                    ga = new GeneticAlgorithm(populationNumber, nnDesign, geneLength, mutationRate);
                    CreateNewFood(foodNumber);
                    CreateNewEntitys(populationNumber);
                    startedSimulation = true;
                    return "Simulation started!";
                }
                else { return "Starting parameters not valid!"; }
            }
            catch (Exception e)
            { return e.Message; }
        }
        public void ResetSimulation()
        {
            sumOfEnergyConsumed = 0;
            numberOfEntityCreated = 0;
            currentGeneration = 0;
            entitysCount = 0;
            food.Clear();
            entitys.Clear();
            ga = new GeneticAlgorithm(populationNumber, nnDesign, geneLength, mutationRate);
            CreateNewFood(foodNumber);
            CreateNewEntitys(populationNumber);
            startedSimulation = true;
            eofTrial = false;
        }
        private KeyboardState CheckKeyboardInput(KeyboardState currentKeyboardState)
        {
            if (currentKeyboardState.IsKeyDown(Keys.Escape)) this.Exit();
            if (currentKeyboardState.IsKeyDown(Keys.F1)) speedOfSimulation = 1;
            if (currentKeyboardState.IsKeyDown(Keys.F2)) speedOfSimulation = 2;
            if (currentKeyboardState.IsKeyDown(Keys.F3)) speedOfSimulation = 3;
            if (currentKeyboardState.IsKeyDown(Keys.F4)) speedOfSimulation = 4;
            if (currentKeyboardState.IsKeyDown(Keys.F5)) speedOfSimulation = 5;
            if (currentKeyboardState.IsKeyDown(Keys.C)) showCromosome = true;
            if (currentKeyboardState.IsKeyDown(Keys.H)) showCromosome = false;
            if (currentKeyboardState.IsKeyDown(Keys.P))
            {
                pausedSimulation = true;
            }
            if (currentKeyboardState.IsKeyDown(Keys.U))
            {
                pausedSimulation = false;
            }
            if (currentKeyboardState.IsKeyDown(Keys.S))
            {
                if (startedSimulation == false)
                {
                    Start(populationNumber, nnDesign, geneLength, mutationRate, neuralNetworkMinWeight, neuralNetworkMaxWeight);
                }
            }
            return currentKeyboardState;
        }
        /// <summary>
        /// Metoda prolazi kroz listu aktivnih entiteta, uklanja jedinke koje su potoile energiju te 
        /// poziva metodu za stvaranje nove jedinke ako ih je manje od inicijalne populacije
        /// </summary>
        private void CheckEntityEnergy()
        {
            List<Entity> entityToRemove = new List<Entity>();
            foreach(Entity e in entitys)
            {
                iterationsInSimulation++;
                e.Energy -= entityTimeEnergyLoss;
                if (e.Energy < 0)
                {
                    entityToRemove.Add(e);
                }
            }
            foreach (Entity e in entityToRemove)
            {
                sumOfEnergyConsumed += e.EnergyConsumed;
                CreateEntityFromParents(entitys[0].Nn.Chromosome, entitys[1].Nn.Chromosome);
                entitys.Remove(e);
                entitysCount++;
                currentGeneration = (int)(entitysCount / populationNumber);
            }
        }
        /// <summary>
        /// Metoda prolazi kroz listu trenutne hrane te uklanja hranu kojoj je energija nula te poziva metodu 
        /// za stvaranje nove hrane ako je ima manje od inicijalno zadane
        /// </summary>
        private void CheckFoodEnergy()
        {
            for (int i = foodNumber - 1; i >= 0; i--)
            {
                iterationsInSimulation++;
                food[i].Energy -= foodEnergyLoss;
                if (food[i].Energy < 0)
                {
                    food.RemoveAt(i);
                    CreateNewFood(1);
                }
            }
        }
        /// <summary>
        /// Metoda provjerava jeli hrana u dovoljnoj blizini jedinke za hranjenje, dodaje energiju hrane jedinki
        /// a postavlja hrani na nulu
        /// </summary>
        /// <param name="e"></param>
        private void EntityFeeding(Entity e)
        {
            for (int i = foodNumber - 1; i >= 0; i--)
            {
                iterationsInSimulation++;
                if (Vector2.Distance(food[i].FoodPosition, e.EntityPosition) <
                   (foodTexture.Width * (food[i].Energy / foodEnergy) * foodScale / 2 + entitetTexture.Width / 2))
                {
                    food[i].Energy -= foodEatingSpeed;
                    e.Energy += foodEatingSpeed;
                    e.EnergyConsumed += food[i].Energy;
                    
                }
            }
            if (e.Energy > entityMaxEnergy) e.Energy -= entityOverFeedingPenalty;
        }
        /// <summary>
        /// Stvara n novih objekata hrane u simulaciji.
        /// </summary>
        /// <param name="number"></param>
        private void CreateNewFood(int n)
        {
            for (int i = 0; i < n; i++)
            {
                randomLocation = new Vector2(rand.Next(GraphicsDevice.Viewport.Width - foodTexture.Width) + foodTexture.Width
                            , rand.Next(GraphicsDevice.Viewport.Height - foodTexture.Height) + foodTexture.Height);
                food.Add(new Food(foodEnergy, randomLocation));
            }
 
        }
        /// <summary>
        /// Stvara n novih objekata Entity u simulaciji.
        /// </summary>
        /// <param name="number"></param>
        /// <param name="entityEnergy"></param>
        /// <param name="rotSpeed"></param>
        /// <param name="forSpeed"></param>
        /// <param name="rotNrgLoss"></param>
        /// <param name="forNrgLoss"></param>
        private void CreateNewEntitys(int n) 
        {
                for (int i = 0; i < n; i++)
                {
                    CreateEntity();
                }
        }
        /// <summary>
        /// Metoda stvara novi objekt Entity po parametrima zadanima simulaciji
        /// i dodaje ga listi entitys.
        /// Neuronska mrea novostvorenog Entity objekta poprima nasumine vrijednosti
        /// </summary>
        /// <returns></returns>
        private void CreateEntity()
        {
            randomLocation = new Vector2(rand.Next(GraphicsDevice.Viewport.Width)
                               , rand.Next(GraphicsDevice.Viewport.Height));
            Entity entityToAdd = new Entity(randomLocation
                                   , rand.Next(360)
                                   , entityEnergy
                                   , entityRotateSpeed
                                   , entityForwardSpeed);
            NeuralNetwork nn = new NeuralNetwork(ga.RandomBitArrayChromosome(), nnDesign, neuralNetworkMinWeight, neuralNetworkMaxWeight);
            entityToAdd.Nn = nn;
            entitys.Add(entityToAdd);
            numberOfEntityCreated++;
                    
        }
        /// <summary>
        /// Metoda stvara novu jedinku koja ima neuronsku mreu nastalu kombinacijom dvaju roditelja
        /// </summary>
        /// <param name="firstParent"></param>
        /// <param name="secondParent"></param>
        /// <returns></returns>
        private void CreateEntityFromParents(List<BitArray[][]> firstParent, List<BitArray[][]> secondParent)
        {
            NeuralNetwork nn = new NeuralNetwork(ga.ChildFromParents(firstParent, secondParent), nnDesign, neuralNetworkMinWeight, neuralNetworkMaxWeight);
            nn.MatrixDesign = nnDesign;

            Vector2 randomLocation = new Vector2(rand.Next(GraphicsDevice.Viewport.Width)
                                                ,rand.Next(GraphicsDevice.Viewport.Height));

            Entity entitiyChild = new Entity(randomLocation
                                   , rand.Next(360)
                                   , entityEnergy
                                   , entityRotateSpeed
                                   , entityForwardSpeed);
            entitiyChild.Nn = nn;
            entitys.Add(entitiyChild);
            numberOfEntityCreated++;
        }
        /// <summary>
        /// Rauna koordinate nove toke udaljene od Vector2 varijable 'position' za 'viewDistance' pod kutom 'angle'.
        /// </summary>
        /// <param name="position"></param>
        /// <param name="angle"></param>
        /// <param name="viewDistance"></param>
        /// <returns></returns>
        public Vector2 CalculateNewPoint(Vector2 position, float angle, double viewDistance)
        {
            Vector2 result = new Vector2();
            result.X = (float)(Math.Cos(MathHelper.ToRadians(angle - 90)) * viewDistance);
            result.Y = (float)(Math.Sin(MathHelper.ToRadians(angle - 90)) * viewDistance);
            result += position;
            return result;
        }
        /// <summary>
        /// Metoda vraa listu double vrijednosti koji reprezentiraju inpute za odreenu lokaciju.
        /// U izraunu se gledaju objekti Entity i Food.
        /// </summary>
        /// <param name="position"></param>
        /// <param name="angle"></param>
        /// <param name="visibilityRange"></param>
        /// <param name="visibilityAngle"></param>
        /// <returns></returns>
        public List<double> CalculateInputs(int index, Vector2 position, float currentAngle)
        {
            double min_distance = visibilityRange;
            double energy_input;
            double threat_input;
            List<double> result = new List<double>();
            foodEyeValues.Clear();
            entityEyeValues.Clear();
            foodEyeValues.Add(0.0f);
            foodEyeValues.Add(0.0f);
            foodEyeValues.Add(0.0f);
            entityEyeValues.Add(0.0f);
            entityEyeValues.Add(0.0f);
            entityEyeValues.Add(0.0f);
            double maxSumFood;
            double maxSumEntity;
            Vector2 directedEyeCenter = CalculateNewPoint(position, currentAngle, visibilityRange);
            Vector2 directedEyeLeft = CalculateNewPoint(position, currentAngle - (visibilityAngle / 2), visibilityRange);
            Vector2 directedEyeRight = CalculateNewPoint(position, currentAngle + (visibilityAngle / 2), visibilityRange);

            for (int i = 0; i < food.Count(); i++)
            {
                if (PointProximityCheck(position, food[i].FoodPosition, visibilityRange))
                {
                    if (Math.Abs(AngleFromLine(position, directedEyeCenter, food[i].FoodPosition)) < visibilityAngle / 2)
                    {
                        foodEyeValues[0] += food[i].Energy * ((double)Math.Pow((visibilityRange - Vector2.Distance(food[i].FoodPosition, position)), 2));
                    }
                    if (Math.Abs(AngleFromLine(position, directedEyeLeft, food[i].FoodPosition)) < visibilityAngle / 2)
                    {
                        foodEyeValues[1] += food[i].Energy * ((double)Math.Pow((visibilityRange - Vector2.Distance(food[i].FoodPosition, position)), 2));
                    }
                    if (Math.Abs(AngleFromLine(position, directedEyeRight, food[i].FoodPosition)) < visibilityAngle / 2)
                    {
                        foodEyeValues[2] += food[i].Energy * ((double)Math.Pow((visibilityRange - Vector2.Distance(food[i].FoodPosition, position)), 2));
                    }
                }
            }

            for (int i = 0; i < entitys.Count(); i++)
            {
                if (position != entitys[i].EntityPosition)
                {
                    if (PointProximityCheck(position, entitys[i].EntityPosition, visibilityRange))
                    {
                        if (Math.Abs(AngleFromLine(position, directedEyeCenter, entitys[i].EntityPosition)) < visibilityAngle / 2)
                        {
                            entityEyeValues[0] += entitys[i].Energy * Vector2.Distance(entitys[i].EntityPosition, position);
                            if (Vector2.Distance(entitys[i].EntityPosition, position) < entitetTexture.Width)
                            {
                                entitys[index].EnergyFromOtherEntitys += entityEatingSpeed;
                                entitys[index].Energy += entityEatingSpeed;
                                entitys[i].Energy -= entityEatingSpeed;
                                entitys[index].EnergyConsumed += entityEatingSpeed;
                            }
                        }
                        if (Math.Abs(AngleFromLine(position, directedEyeLeft, entitys[i].EntityPosition)) < visibilityAngle / 2)
                        {
                            entityEyeValues[1] += entitys[i].Energy * Vector2.Distance(entitys[i].EntityPosition, position);
                        }
                        if (Math.Abs(AngleFromLine(position, directedEyeRight, entitys[i].EntityPosition)) < visibilityAngle / 2)
                        {
                            entityEyeValues[2] += entitys[i].Energy * Vector2.Distance(entitys[i].EntityPosition, position); 
                        }
                    }
                }
            }
            foreach (Entity e in entitys)
            {
                double distance = Vector2.Distance(position, e.EntityPosition);
                if (index != entitys.IndexOf(e))
                {
                    if (min_distance > distance)
                    {
                        min_distance = distance;
                    }   
                }
            }
            threat_input = (visibilityRange - min_distance) / visibilityRange;
            maxSumFood = foodEyeValues.Max();
            maxSumEntity = entityEyeValues.Max();

            if (maxSumFood == 0)
            {
                maxSumFood = 1;
                for (int i = 0; i < foodEyeValues.Count(); i++)
                {
                    foodEyeValues[i] = 1f;
                }
            }
            for (int i = 0; i < foodEyeValues.Count; i++)
            {
                foodEyeValues[i] = foodEyeValues[i] / maxSumFood;
                result.Add(foodEyeValues[i]);
            }
            
            if (maxSumEntity == 0)
            {
                maxSumEntity = 1f;
                for (int i = 0; i < entityEyeValues.Count(); i++)
                {
                    entityEyeValues[i] = 0.0f;
                }
            }
            for (int i = 0; i < entityEyeValues.Count; i++)
            {
                entityEyeValues[i] = entityEyeValues[i] / maxSumEntity;
                result.Add(entityEyeValues[i]);
            }
            if (entitys[index].Energy < 255)
            {
                energy_input = entitys[index].Energy / entityMaxEnergy;
            }
            else { energy_input = 1; }
            result.Add(threat_input);
            result.Add(energy_input);
            return result;

        } 
        /// <summary>
        ///  Funkcija vraa true ako je udaljenost dvije toke manja od ulaznog parametra 'distance'.
        /// </summary>
        /// <param name="firstPoint"></param>
        /// <param name="secondPoint"></param>
        /// <param name="distance"></param>
        /// <returns></returns>
        public bool PointProximityCheck(Vector2 firstPoint, Vector2 secondPoint, double distance)
        {
            if (Vector2.Distance(firstPoint, secondPoint) < distance)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        /// <summary>
        /// Metoda rauna kut izmeu duina firstLinePoint-secondLinePoint i firstLinePoint-point
        /// </summary>
        /// <param name="firstLinePoint"></param>
        /// <param name="secondLinePoint"></param>
        /// <param name="point"></param>
        /// <returns></returns>
        public float AngleFromLine(Vector2 firstLinePoint, Vector2 secondLinePoint, Vector2 point)
        {
            return MathHelper.ToDegrees((float)Math.Atan2(secondLinePoint.Y - firstLinePoint.Y, secondLinePoint.X - firstLinePoint.X)- (float)Math.Atan2(point.Y - firstLinePoint.Y, point.X - firstLinePoint.X));
        }
        /// <summary>
        /// Metoda se koristi za promjenu trenutne pozicije entiteta ako je izaao van granica simulacije.
        /// </summary>
        /// <param name="e"></param>
        /// <returns></returns>
        public Vector2 CalculatePositionAdjustment(Vector2 currentPosition)
        {
            Vector2 result = currentPosition;
            if (currentPosition.X < 0) result.X = GraphicsDevice.Viewport.Width;
            if ((currentPosition.X) > GraphicsDevice.Viewport.Width) result.X = 0;
            if (currentPosition.Y < 0) result.Y = GraphicsDevice.Viewport.Height;
            if ((currentPosition.Y) > GraphicsDevice.Viewport.Height) result.Y = 0;
            return result;
        }

        public void CreateNewGeneration(int populationNumber, List<BitArray[][]> firstParent, List<BitArray[][]> secondParent)
        {
            currentGeneration++;
            for(int i = 0; i < populationNumber; i++)
            { CreateEntityFromParents(firstParent, secondParent); }
        }
        #endregion

        #region metode grafike
        public Color CalculateOverlay(double value, double referenceValue, double ratio)
        {
            int overlayRed, overlayGreen, overlayBlue, overlayAlpha;
                
            overlayRed = (int)ratio * 255;
            overlayGreen = (int)(1 - ratio) * 255;
        
            if (value > referenceValue)
            { 
                overlayBlue = 255;
                overlayAlpha = 255;
            }
            else 
            { 
                overlayBlue = (int)((value / referenceValue) * 255);
                overlayAlpha = (int)Math.Pow(value, 2);
            }

            Color result = new Color(overlayRed, overlayBlue, overlayGreen, overlayAlpha);
            return result;
        }

        public void DrawStrings(List<string> strings, SpriteBatch sprBatch, Vector2 startLocation, SpriteFont spriteFont, Color color)
        {
            Vector2 size;
            Rectangle rectangle;
            Color colori = new Color(205, 201, 201, 115);
            for (int i = 0; i < strings.Count; i++)
            {
                size = spriteFont.MeasureString(strings[i]);
                
                rectangle = new Rectangle((int)startLocation.X, (int)(startLocation.Y + spriteFont.LineSpacing*i), (int)size.X, (int)size.Y);
                sprBatch.Draw(backTexture, rectangle, colori);
                sprBatch.DrawString(spriteFont, strings[i], new Vector2(startLocation.X, startLocation.Y + (spriteFont.LineSpacing*i)), color);
            }
        }

        public void DrawChromosome(SpriteBatch sprBatch, List<BitArray[][]> chromosome, Vector2 startLocation, float drawSize)
        {
            Rectangle location;
            float matrixOffset = 0;
            int alphaValue;
            for (int i = 0; i < nnDesign.Count() - 1; i++)
            {
                
                BitArray[][] currentMatrix = chromosome[i];

                for (int j = 0; j < nnDesign[i]; j++)
                {
                    
                    for (int k = 0; k < nnDesign[i + 1]; k++)
                    {
                        alphaValue = 255;
                        location = new Rectangle((int)(startLocation.X + k*drawSize), (int)(startLocation.Y + matrixOffset*drawSize + j*drawSize), (int)drawSize, (int)drawSize);
                        byte[] Colors = GetColorsFromBitArray(currentMatrix[j][k]);
                        if (currentMatrix[j][k][geneLength - 1] == true) alphaValue = 128;
                        sprBatch.Draw(backTexture, location, new Color(Colors[0], Colors[1], Colors[2], alphaValue)); 
                    }
                }
                matrixOffset += nnDesign[i];
            }
        }


        private byte[] GetColorsFromBitArray(BitArray bitArray)
        {
            byte[] array = new byte[4];
            bitArray.CopyTo(array, 0);
            return array;
        }


        #endregion

        #region save/load metode
        public void Save(string location)
        {
            BinaryWriter w = new BinaryWriter(File.Open(location + ".parameters.bin", FileMode.Create));
            //Spremanje varijabli entity-a
            w.Write(entityEnergy);
            w.Write(entityMaxEnergy);
            w.Write(entityOverFeedingPenalty);
            w.Write(entityForwardSpeed);
            w.Write(entityRotateSpeed);
            w.Write(entityTimeEnergyLoss);
            w.Write(visibilityRange);
            w.Write(visibilityAngle);
            w.Write(foodEatingSpeed);
            w.Write(entityEatingSpeed);

            //Spremanje varijabli hrane
            w.Write(foodNumber);
            w.Write(foodEnergy);
            w.Write(foodEnergyLoss);
            w.Write(foodScale);

            //Spremanje varijabli NN-a
            w.Write(nnDesign.Count());
            for(int i = 0; i< nnDesign.Count(); i++)
            { w.Write(nnDesign[i]);}
            w.Write(neuralNetworkMaxWeight);
            w.Write(neuralNetworkMinWeight);

            //Spremanje varijabli GA
            w.Write(populationNumber);
            w.Write(geneLength);
            w.Write(mutationRate);

            SavePopulation(location);
            SaveFood(location);
        }

        public void SavePopulation(string location)
        {

                BinaryWriter w = new BinaryWriter(File.Open(location + ".entitys.bin", FileMode.Create));
                foreach (Entity e in entitys)
                {
                    for (int i = 0; i < nnDesign.Count() - 1; i++)
                    {
                        for (int j = 0; j < nnDesign[i]; j++)
                        {
                            for (int k = 0; k < nnDesign[i + 1]; k++)
                            {
                                w.Write(getIntFromBitArray(e.Nn.Chromosome[i][j][k]));
                            }

                        }
                    }
                    w.Write(e.EntityPosition.X);
                    w.Write(e.EntityPosition.Y);
                    w.Write(e.AngleOfMovement);
                }
                w.Close();
        }

        public string LoadPopulation(string location, int number)
        {
            try
            {
                FileStream fs = new FileStream(location + ".entitys.bin", FileMode.Open, FileAccess.Read);
                BinaryReader r = new BinaryReader(fs);
                entitys.Clear();
                if (number > populationNumber) number = populationNumber;
                if (number < 3) number = 3;
                for (int z = 0; z < number; z++)
                {
                    List<BitArray[][]> savedChromosome = new List<BitArray[][]>();
                    for (int i = 0; i < nnDesign.Count() - 1; i++)
                    {
                        BitArray[][] matrixToAdd = new BitArray[nnDesign[i]][];

                        for (int j = 0; j < nnDesign[i]; j++)
                        {
                            matrixToAdd[j] = new BitArray[nnDesign[i + 1]];

                            for (int k = 0; k < nnDesign[i + 1]; k++)
                            {
                                BitArray b = new BitArray(new int[] { r.ReadInt32() });
                                b.Length = geneLength;
                                matrixToAdd[j][k] = b;
                            }
                        }
                        savedChromosome.Add(matrixToAdd);
                    }
                    Entity entityToAdd = new Entity(new Vector2(r.ReadSingle(), r.ReadSingle())
                                           , r.ReadSingle()
                                           , entityEnergy
                                           , entityRotateSpeed
                                           , entityForwardSpeed);
                    entityToAdd.Nn = new NeuralNetwork(savedChromosome, nnDesign, neuralNetworkMinWeight, neuralNetworkMaxWeight);
                    entitys.Add(entityToAdd);
                }
                r.Close();
                fs.Close();
                return "Entity loading - Success! ";
            }
            catch (Exception e)
            {
                return e.Message;
            }

        }

        public string Load(string location, int number)
        {
            try
            {
                FileStream fs = new FileStream(location + ".parameters.bin", FileMode.Open, FileAccess.Read);
                BinaryReader r = new BinaryReader(fs);
                //Uitavanje varijabli entiteta
                entityEnergy = r.ReadDouble();
                entityMaxEnergy = r.ReadDouble();
                entityOverFeedingPenalty = r.ReadDouble();
                entityForwardSpeed = r.ReadDouble();
                entityRotateSpeed = r.ReadDouble();
                entityTimeEnergyLoss = r.ReadDouble();
                visibilityRange = r.ReadDouble();
                visibilityAngle = r.ReadSingle();
                foodEatingSpeed = r.ReadDouble();
                entityEatingSpeed = r.ReadDouble();


                //Uitavanje varijabli hrane
                foodNumber = r.ReadInt32();
                foodEnergy = r.ReadDouble();
                foodEnergyLoss = r.ReadDouble();
                foodScale = r.ReadSingle();

                //Uitavanje varijabli NN-a
                int counter = r.ReadInt32();
                nnDesign.Clear();
                for (int i = 0; i < counter; i++)
                { nnDesign.Add(r.ReadInt32()); }
                neuralNetworkMaxWeight = r.ReadDouble();
                neuralNetworkMinWeight = r.ReadDouble();

                //Uitavanje varijabli GA
                populationNumber = r.ReadInt32();
                geneLength = r.ReadInt32();
                mutationRate = r.ReadDouble();
                r.Close();
                fs.Close();

                sumOfEnergyConsumed = 0;
                currentGeneration = 0;
                entitysCount = 0;
                food.Clear();
                entitys.Clear();
                ga = new GeneticAlgorithm(populationNumber, nnDesign, geneLength, mutationRate);
                CreateNewFood(foodNumber);
                startedSimulation = true;
                return LoadPopulation(location, number) + LoadFood(location);
            }
            catch (Exception e)
            { return e.Message; }
        }

        public void SaveFood(string location)
        {
            BinaryWriter w = new BinaryWriter(File.Open(location + ".food.bin", FileMode.Create));
            foreach (Food f in food)
            {
                w.Write(f.Energy);
                w.Write(f.FoodPosition.X);
                w.Write(f.FoodPosition.Y);
            }
            w.Close();
        }

        public string LoadFood(string location)
        {
            try
            {
                FileStream fs = new FileStream(location + ".food.bin", FileMode.Open, FileAccess.Read);
                BinaryReader r = new BinaryReader(fs);
                food.Clear();
                for(int i = 0; i<foodNumber; i++)
                {
                    food.Add(new Food(r.ReadDouble(), new Vector2(r.ReadSingle(), r.ReadSingle())));
                }
                fs.Close();
                r.Close();
                return "Food loading - Success!"; 
            }
            catch (Exception e)
            { return e.Message;  }
        }

        static int getIntFromBitArray(BitArray bitArray)
        {
            int[] array = new int[1];
            bitArray.CopyTo(array, 0);
            return array[0];
        }

        #endregion

    }
}

